# Copyright (C) 2018  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import urllib.parse


import bottle


import qvisqve
import qvisqve_secrets


class TokenRouter(qvisqve.Router):

    def __init__(self, token_generator, clients):
        super().__init__()
        self._generator = token_generator
        self._clients = Clients(clients)

    def get_routes(self):
        return [
            {
                'method': 'POST',
                'path': '/token',
                'callback': self._create_token,
                'needs-authorization': False,
            },
        ]

    def _create_token(self, content_type, body, **kwargs):
        qvisqve.log.log('xxx', body=body, kwargs=kwargs)

        if content_type != 'application/x-www-form-urlencoded':
            return qvisqve.bad_request_response('Wrong content type')

        client_id, client_secret = bottle.request.auth
        if not self._clients.is_correct_secret(client_id, client_secret):
            return qvisqve.unauthorized_response('Unauthorized')

        params = self._get_form_params(body)

        grant_type = self._get_grant_type(params)
        if grant_type != 'client_credentials':
            return qvisqve.bad_request_response('Wrong grant type')

        scope = self._get_scope(params)
        if scope is None:
            return qvisqve.bad_request_response('Bad scope')

        allowed = self._clients.get_allowed_scopes(client_id)
        scope = ' '.join(
            s
            for s in scope.split()
            if s in allowed
        )

        token = self._generator.new_token(client_id, scope)
        return qvisqve.ok_response({
            'access_token': token,
            'token_type': 'bearer',
            'scope': scope,
        })

    def _get_form_params(self, body):
        body = body.decode('UTF-8')
        return urllib.parse.parse_qs(body)

    def _get_grant_type(self, params):
        grant_type = params.get('grant_type')
        if len(grant_type) == 1:
            return grant_type[0]
        return None

    def _get_scope(self, params):
        scope = params.get('scope', [])
        if len(scope) > 1:
            return None
        if scope:
            return scope[0]
        return ''


class Clients:

    def __init__(self, clients):
        self._clients = clients
        self._hasher = qvisqve_secrets.SecretHasher()

    def is_correct_secret(self, client_id, cleartext):
        client = self._get_client(client_id)
        secret = client.get('client_secret')
        return secret and self._hasher.is_correct(secret, cleartext)

    def get_allowed_scopes(self, client_id):
        client = self._get_client(client_id)
        return client.get('allowed_scopes', [])

    def _get_client(self, client_id):
        return self._clients.get(client_id, {})
